"""loader
flattened XML specific

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Optional, Union
from json import dumps
from os.path import isfile, basename

from .i18n import _
from .error import DictConsistencyError, VariableCalculationDependencyError
from .utils import normalize_family
from .object_model import Calculation, CONVERT_OPTION


class BaseElt:  # pylint: disable=R0903
    """Base element"""

    path = "."
    type = "family"


def sorted_func_name(func_name):
    s_func_name = func_name.split("/")
    s_func_name.reverse()
    return "/".join(s_func_name)


class TiramisuReflector:
    """Convert object to tiramisu representation"""

    def __init__(
        self,
        objectspace,
        funcs_paths,
    ):
        self.informations_idx = -1
        self.reflector_objects = {}
        self.text = {
            "header": [],
            "option": [],
        }
        self.objectspace = objectspace
        if self.objectspace.export_with_import:
            if self.objectspace.internal_functions:
                for func in self.objectspace.internal_functions:
                    self.text["header"].append(f"func[func] = func")
            self.text["header"].extend(
                [
                    "from tiramisu import *",
                    "from tiramisu.setting import ALLOWED_LEADER_PROPERTIES",
                    "from re import compile as re_compile",
                ]
            )
        if self.objectspace.export_with_import:
            self.text["header"].extend(
                [
                    "from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription"
                ]
            )
        if funcs_paths:
            for funcs_path in sorted(funcs_paths, key=sorted_func_name):
                if not isfile(funcs_path):
                    continue
                self.text["header"].append(f"load_functions('{funcs_path}')")
        if self.objectspace.export_with_import:
            if objectspace.main_namespace:
                self.text["header"].extend(
                    [
                        "try:",
                        "    groups.namespace",
                        "except:",
                        "    groups.addgroup('namespace')",
                    ]
                )
            for mode in self.objectspace.modes_level:
                self.text["header"].append(f'ALLOWED_LEADER_PROPERTIES.add("{mode}")')
        self.make_tiramisu_objects()
        for key, value in self.objectspace.jinja.items():
            self.add_jinja_to_function(key, value)

    def add_jinja_to_function(
        self,
        variable_name: str,
        jinja: str,
    ) -> None:
        jinja_text = dumps(jinja, ensure_ascii=False)
        self.text["header"].append(f"dict_env['{variable_name}'] = {jinja_text}")

    def make_tiramisu_objects(self) -> None:
        """make tiramisu objects"""
        baseelt = BaseElt()
        self.objectspace.reflector_names[baseelt.path] = (
            f"option_0{self.objectspace.suffix}"
        )
        basefamily = Family(
            baseelt,
            self,
        )
        for elt in self.objectspace.paths.get():
            if elt.path in self.objectspace.families:
                Family(
                    elt,
                    self,
                )
            else:
                Variable(
                    elt,
                    self,
                )
        #        else:
        #            path_prefixes = self.objectspace.paths.get_path_prefixes()
        #            for path_prefix in path_prefixes:
        #                space = self.objectspace.space.variables[path_prefix]
        #                self.set_name(space)
        #                baseprefix = Family(
        #                    space,
        #                    self,
        #                )
        #                basefamily.add(baseprefix)
        #                for elt in self.reorder_family(space):
        #                    self.populate_family(
        #                        baseprefix,
        #                        elt,
        #                    )
        #                if not hasattr(baseprefix.elt, "information"):
        #                    baseprefix.elt.information = self.objectspace.information(
        #                        baseprefix.elt.xmlfiles
        #                    )
        #                for key, value in self.objectspace.paths.get_providers_path(
        #                    path_prefix
        #                ).items():
        #                    setattr(baseprefix.elt.information, key, value)
        #                for key, value in self.objectspace.paths.get_suppliers_path(
        #                    path_prefix
        #                ).items():
        #                    setattr(baseprefix.elt.information, key, value)
        baseelt.name = normalize_family(self.objectspace.base_option_name)
        baseelt.description = self.objectspace.base_option_name
        self.reflector_objects[baseelt.path].get(
            [], baseelt.description
        )  # pylint: disable=E1101

    def set_name(
        self,
        elt,
    ):
        """Set name"""
        if elt.path not in self.objectspace.reflector_names:
            self.objectspace.set_name(elt, "optiondescription_")
        return self.objectspace.reflector_names[elt.path]

    def get_information_name(self):
        self.informations_idx += 1
        return f"information_{self.informations_idx}"

    def get_text(self):
        """Get text"""
        return "\n".join(self.text["header"] + self.text["option"])


class Common:
    """Common function for variable and family"""

    def __init__(
        self,
        elt,
        tiramisu,
    ):
        self.objectspace = tiramisu.objectspace
        self.elt = elt
        self.option_name = None
        self.tiramisu = tiramisu
        tiramisu.reflector_objects[elt.path] = self
        self.object_type = None
        self.informations = []

    def get(self, calls, parent_name):
        """Get tiramisu's object"""
        if self.elt.path in calls:
            msg = f'"{self.elt.path}" will make an infinite loop'
            raise DictConsistencyError(msg, 80, self.elt.xmlfiles)
        self_calls = calls.copy()
        self_calls.append(self.elt.path)
        self.calls = self_calls
        if self.option_name is None:
            self.option_name = self.objectspace.reflector_names[self.elt.path]
            self.populate_attrib()
            if self.informations:
                for information in self.informations:
                    self.tiramisu.text["option"].append(
                        f"{information}.set_option({self.option_name})"
                    )
        return self.option_name

    def populate_attrib(self):
        """Populate attributes"""
        keys = {"name": self.convert_str(self.elt.name)}
        if hasattr(self.elt, "description") and self.elt.description:
            keys["doc"] = self.convert_str(self.elt.description)
        self._populate_attrib(keys)
        if self.elt.path in self.objectspace.properties:
            keys["properties"] = self.properties_to_string(
                self.objectspace.properties[self.elt.path]
            )
        self.populate_informations(keys)
        attrib = ", ".join([f"{key}={value}" for key, value in keys.items()])
        self.tiramisu.text["option"].append(
            f"{self.option_name} = {self.object_type}({attrib})"
        )

    def _populate_attrib(
        self,
        keys: dict,
    ) -> None:  # pragma: no cover
        raise NotImplementedError()

    @staticmethod
    def convert_str(value):
        """convert string"""
        if value is None:
            return "None"
        return dumps(value, ensure_ascii=False)

    def properties_to_string(
        self,
        values: list,
    ) -> None:
        """Change properties to string"""
        properties = []
        calc_properties = []
        for property_, value in values.items():
            if value is True:
                properties.append(self.convert_str(property_))
            elif isinstance(value, list):
                for val in value:
                    calc_properties.append(self.calculation_value(val))
            else:
                calc_properties.append(self.calculation_value(value))
        return "frozenset({" + ", ".join(sorted(properties) + calc_properties) + "})"

    def calc_properties(
        self,
        prop,
        calculation,
    ) -> str:
        """Populate properties"""
        option_name = self.tiramisu.reflector_objects[child.source.path].get(
            self.calls, self.elt.path
        )
        kwargs = (
            f"'condition': ParamOption({option_name}, notraisepropertyerror=True), "
            f"'expected': {self.populate_param(child.expected)}"
        )
        if child.inverse:
            kwargs += ", 'reverse_condition': ParamValue(True)"
        return (
            f"Calculation(func['calc_value'], Params(ParamValue('{child.name}'), "
            f"kwargs={{{kwargs}}}), func['calc_value_property_help'])"
        )

    def populate_informations(self, keys):
        """Populate Tiramisu's informations"""
        informations = self.objectspace.informations.get(self.elt.path)
        if not informations:
            return
        keys["informations"] = informations

    def populate_param(
        self,
        param,
    ):
        """Populate variable parameters"""
        if not isinstance(param, dict):
            if isinstance(param, str):
                value = self.convert_str(param)
            else:
                value = param
            return f"ParamValue({value})"
        if param["type"] == "value":
            return f"ParamValue({param['value']})"
        if param["type"] == "information":
            # default? really?
            if self.elt.multi:
                default = []
            else:
                default = None
            if "variable" in param:
                if param["variable"].path == self.elt.path:
                    return f'ParamSelfInformation("{param["information"]}", {default})'
                information_variable_path = param["variable"].path
                information_variable = self.tiramisu.reflector_objects[
                    information_variable_path
                ]
                if information_variable_path not in self.calls:
                    option_name = information_variable.get(self.calls, self.elt.path)
                    return f'ParamInformation("{param["information"]}", {default}, option={option_name})'
                else:
                    information = (
                        f'ParamInformation("{param["information"]}", {default})'
                    )
                    information_name = self.tiramisu.get_information_name()
                    self.tiramisu.text["option"].append(
                        f"{information_name} = {information}"
                    )
                    information_variable.informations.append(information_name)
                    return information_name
            return f'ParamInformation("{param["information"]}", {default})'
        if param["type"] == "identifier":
            if "identifier" in param and param["identifier"] != None:
                return f"ParamIdentifier(identifier_index={param['identifier']})"
            return "ParamIdentifier()"
        if param["type"] == "index":
            return "ParamIndex()"
        if param["type"] == "variable":
            return self.build_option_param(
                param["variable"],
                param.get("propertyerror", True),
                param.get("identifier"),
                param.get("dynamic"),
                param.get("whole", False),
            )
        if param["type"] == "any":
            if isinstance(param["value"], str):
                value = self.convert_str(param["value"])
            else:
                value = str(param["value"])
            return "ParamValue(" + value + ")"
        raise Exception("pfff")

    def build_option_param(
        self,
        variable,
        propertyerror,
        identifier: Optional[str],
        dynamic,
        whole: bool,
    ) -> str:
        """build variable parameters"""
        if variable.path == self.elt.path:
            return f"ParamSelfOption(whole={whole})"
        if whole:
            msg = f'variable param "{variable.path}" has whole attribute but it\'s not allowed for external variable'
            raise DictConsistencyError(msg, 34, self.elt.xmlfiles)
        option_name = self.tiramisu.reflector_objects[variable.path].get(
            self.calls, self.elt.path
        )
        params = [f"{option_name}"]
        if identifier is not None:
            param_type = "ParamDynOption"
            identifiers = []
            for ident in identifier:
                if isinstance(ident, str):
                    ident = self.convert_str(ident)
                identifiers.append(str(ident))
            params.append("[" + ", ".join(identifiers) + "]")
        else:
            param_type = "ParamOption"
        if not propertyerror:
            params.append("notraisepropertyerror=True")
        return f'{param_type}({", ".join(params)})'

    def calculation_value(
        self,
        function,
    ) -> str:
        """Generate calculated value"""
        child = function.to_function(self.objectspace)
        new_args = []
        kwargs = []
        if "params" in child:
            for key, value in child["params"].items():
                if not key:
                    for val in value:
                        new_args.append(self.populate_param(val))
                else:
                    kwargs.append(f"'{key}': " + self.populate_param(value))
        ret = (
            f"Calculation(func['{child['function']}'], Params(("
            + ", ".join(new_args)
            + ")"
        )
        if kwargs:
            ret += ", kwargs={" + ", ".join(kwargs) + "}"
        ret += ")"
        if hasattr(child, "warnings_only"):
            ret += f", warnings_only={child.warnings_only}"
        if "help" in child:
            ret += f", help_function=func['{child['help']}']"
        ret = ret + ")"
        return ret

    def populate_calculation(
        self,
        datas: Union[Calculation, str, list],
        return_a_tuple: bool = False,
    ) -> str:
        if isinstance(datas, str):
            return self.convert_str(datas)
        if isinstance(datas, Calculation):
            return self.calculation_value(datas)
        if not isinstance(datas, list):
            return datas
        params = []
        for idx, data in enumerate(datas):
            if isinstance(data, Calculation):
                try:
                    params.append(self.calculation_value(data))
                except VariableCalculationDependencyError:
                    pass
            elif isinstance(data, str):
                params.append(self.convert_str(data))
            else:
                params.append(str(data))
        if return_a_tuple:
            ret = "("
        else:
            ret = "["
        ret += ", ".join(params)
        if return_a_tuple:
            if len(params) <= 1:
                ret += ","
            ret += ")"
        else:
            ret += "]"
        return ret


class Variable(Common):
    """Manage variable"""

    def __init__(
        self,
        elt,
        tiramisu,
    ):
        super().__init__(elt, tiramisu)
        if elt.type in self.tiramisu.objectspace.custom_types:
            self.object_type = self.tiramisu.objectspace.custom_types[elt.type].__name__
        else:
            self.object_type = CONVERT_OPTION[elt.type]["opttype"]

    def _populate_attrib(
        self,
        keys: dict,
    ):
        if self.elt.type == "symlink":
            keys["opt"] = self.tiramisu.reflector_objects[self.elt.opt.path].get(
                self.calls, self.elt.path
            )
            return
        if self.elt.type == "choice":
            keys["values"] = self.populate_calculation(
                self.elt.choices, return_a_tuple=True
            )
        if self.elt.type == "regexp":
            self.object_type = "Regexp_" + self.option_name
            self.tiramisu.text["header"].append(
                f"""class {self.object_type}(RegexpOption):
    __slots__ = tuple()
    _type = 'value'
{self.object_type}._regexp = re_compile(r"{self.elt.regexp}")
"""
            )
        if self.elt.path in self.objectspace.multis:
            keys["multi"] = self.objectspace.multis[self.elt.path]
        if hasattr(self.elt, "default") and self.elt.default is not None:
            try:
                keys["default"] = self.populate_calculation(self.elt.default)
            except VariableCalculationDependencyError:
                pass
        if self.elt.path in self.objectspace.default_multi:
            try:
                keys["default_multi"] = self.populate_calculation(
                    self.objectspace.default_multi[self.elt.path]
                )
            except VariableCalculationDependencyError:
                pass
        if self.elt.validators:
            keys["validators"] = self.populate_calculation(self.elt.validators)
        for key, value in (
            CONVERT_OPTION.get(self.elt.type, {}).get("initkwargs", {}).items()
        ):
            if isinstance(value, str):
                value = self.convert_str(value)
            keys[key] = value
        if self.elt.params:
            for param in self.elt.params:
                value = param.value
                if isinstance(value, str):
                    value = self.convert_str(value)
                keys[param.key] = value


class Family(Common):
    """Manage family"""

    def __init__(
        self,
        elt,
        tiramisu,
    ):
        super().__init__(elt, tiramisu)
        if self.elt.type == "dynamic":
            self.tiramisu.objectspace.has_dyn_option = True
            self.object_type = "ConvertDynOptionDescription"
        elif self.elt.type == "leadership":
            self.object_type = "Leadership"
        else:
            self.object_type = "OptionDescription"
        if hasattr(self.elt, "name") and self.elt.name == self.elt.namespace:
            self.group_type = "groups.namespace"
        else:
            self.group_type = None
        self.children = []

    def add(self, child):
        """Add a child"""
        self.children.append(child)

    def _populate_attrib(
        self,
        keys: list,
    ) -> None:
        if self.group_type:
            keys["group_type"] = self.group_type
        if self.elt.type == "dynamic":
            keys["identifiers"] = self.populate_calculation(self.elt.dynamic)
        children = []
        for path in self.objectspace.parents[self.elt.path]:
            children.append(self.objectspace.paths[path])
        keys["children"] = (
            "["
            + ", ".join(
                [
                    self.tiramisu.reflector_objects[child.path].get(
                        self.calls, self.elt.path
                    )
                    for child in children
                ]
            )
            + "]"
        )



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_1 = BoolOption(name="condition", doc="a condition", default=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_2 = StrOption(name="variable", doc="a variable", properties=frozenset({"basic", "mandatory", Calculation(func['variable_to_property'], Params((ParamValue("disabled"), ParamOption(option_1)), kwargs={'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['variable_to_property'])}), informations={'type': 'string'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, option_2])



"""Rougail object model

Silique (https://www.silique.fr)
Copyright (C) 2023-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Optional, Union, get_type_hints, Any, Literal, List, Dict, Iterator
from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictFloat,
    StrictStr,
    ConfigDict,
)
from tiramisu import undefined
from .utils import get_jinja_variable_to_param, get_realpath
from .error import DictConsistencyError, VariableCalculationDependencyError

BASETYPE = Union[StrictBool, StrictInt, StrictFloat, StrictStr, None]
PROPERTY_ATTRIBUTE = ["frozen", "hidden", "disabled", "mandatory"]


def convert_boolean(value: str) -> bool:
    """Boolean coercion. The Rougail XML may contain srings like `True` or `False`"""
    if isinstance(value, bool):
        return value
    value = value.lower()
    if value == "true":
        return True
    elif value == "false":
        return False
    elif value in ["", None]:
        return None
    raise Exception(f'unknown boolean value "{value}"')


CONVERT_OPTION = {
    "string": dict(opttype="StrOption", example="example"),
    "number": dict(opttype="IntOption", func=int, example=42),
    "float": dict(opttype="FloatOption", func=float, example=1.42),
    "boolean": dict(opttype="BoolOption", func=convert_boolean),
    "secret": dict(opttype="PasswordOption", example="secrets"),
    "mail": dict(opttype="EmailOption", example="user@example.net"),
    "unix_filename": dict(opttype="FilenameOption", example="/tmp/myfile.txt"),
    "date": dict(opttype="DateOption", example="2000-01-01"),
    "unix_user": dict(opttype="UsernameOption", example="username"),
    "ip": dict(
        opttype="IPOption", initkwargs={"allow_reserved": True}, example="1.1.1.1"
    ),
    "cidr": dict(opttype="IPOption", initkwargs={"cidr": True}, example="1.1.1.0/24"),
    "netmask": dict(opttype="NetmaskOption", example="255.255.255.0"),
    "network": dict(opttype="NetworkOption", example="1.1.1.0"),
    "network_cidr": dict(
        opttype="NetworkOption", initkwargs={"cidr": True}, example="1.1.1.0/24"
    ),
    "broadcast": dict(opttype="BroadcastOption", example="1.1.1.255"),
    "netbios": dict(
        opttype="DomainnameOption",
        initkwargs={"type": "netbios", "warnings_only": True},
        example="example",
    ),
    "domainname": dict(
        opttype="DomainnameOption",
        initkwargs={"type": "domainname", "allow_ip": False},
        example="example.net",
    ),
    "hostname": dict(
        opttype="DomainnameOption",
        initkwargs={"type": "hostname", "allow_ip": False},
        example="example",
    ),
    "web_address": dict(
        opttype="URLOption",
        initkwargs={"allow_ip": False, "allow_without_dot": True},
        example="https://example.net",
    ),
    "port": dict(
        opttype="PortOption", initkwargs={"allow_private": True}, example="111"
    ),
    "mac": dict(opttype="MACOption", example="00:00:00:00:00"),
    "unix_permissions": dict(
        opttype="PermissionsOption",
        initkwargs={"warnings_only": True},
        func=int,
        example="644",
    ),
    "choice": dict(opttype="ChoiceOption", example="a_choice"),
    "regexp": dict(opttype="RegexpOption"),
    #
    "symlink": dict(opttype="SymLinkOption"),
}


class Param(BaseModel):
    key: str
    model_config = ConfigDict(extra="forbid")

    def __init__(
        self,
        path,
        attribute,
        family_is_dynamic,
        is_follower,
        xmlfiles,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)


class AnyParam(Param):
    type: str
    value: Union[BASETYPE, List[BASETYPE]]


class VariableParam(Param):
    type: str
    variable: str
    propertyerror: bool = True
    whole: bool = False
    optional: bool = False


class IdentifierParam(Param):
    type: str
    identifier: Optional[int] = None

    def __init__(
        self,
        **kwargs,
    ) -> None:
        if not kwargs["family_is_dynamic"]:
            msg = f'identifier parameter for "{kwargs["attribute"]}" in "{kwargs["path"]}" cannot be set none dynamic family'
            raise DictConsistencyError(msg, 10, kwargs["xmlfiles"])
        super().__init__(**kwargs)


class InformationParam(Param):
    type: str
    information: str
    variable: Optional[str] = None


class IndexParam(Param):
    type: str

    def __init__(
        self,
        **kwargs,
    ) -> None:

        if not kwargs["is_follower"]:
            msg = f'the variable "{kwargs["path"]}" is not a follower, so cannot have index type for param in "{kwargs["attribute"]}"'
            raise DictConsistencyError(msg, 25, kwargs["xmlfiles"])
        super().__init__(**kwargs)


PARAM_TYPES = {
    "any": AnyParam,
    "variable": VariableParam,
    "identifier": IdentifierParam,
    "information": InformationParam,
    "index": IndexParam,
}


class Calculation(BaseModel):
    path_prefix: Optional[str]
    path: str
    inside_list: bool
    version: str
    ori_path: Optional[str] = None
    default_values: Any = None
    namespace: Optional[str]
    xmlfiles: List[str]

    model_config = ConfigDict(extra="forbid")

    def get_realpath(
        self,
        path: str,
    ) -> str:
        return get_realpath(path, self.path_prefix)

    def get_params(self, objectspace):
        if not self.params:
            return {}
        params = {}
        for param_obj in self.params:
            param = param_obj.model_dump()
            if param.get("type") == "variable":
                if self.ori_path is None:
                    path = self.path
                else:
                    path = self.ori_path
                variable, identifier = objectspace.paths.get_with_dynamic(
                    param["variable"],
                    self.path_prefix,
                    path,
                    self.version,
                    self.namespace,
                    self.xmlfiles,
                )
                if not variable:
                    if not param.get("optional"):
                        msg = f'cannot find variable "{param["variable"]}" defined attribute in "{self.attribute_name}" for "{self.path}"'
                        raise DictConsistencyError(msg, 22, self.xmlfiles)
                    continue
                if not isinstance(variable, objectspace.variable):
                    raise Exception("pfff it's a family")
                param["variable"] = variable
                if identifier:
                    param["identifier"] = identifier
            if param.get("type") == "information":
                if param["variable"]:
                    if self.ori_path is None:
                        path = self.path
                    else:
                        path = self.ori_path
                    variable, identifier = objectspace.paths.get_with_dynamic(
                        param["variable"],
                        self.path_prefix,
                        path,
                        self.version,
                        self.namespace,
                        self.xmlfiles,
                    )
                    if not variable:
                        msg = f'cannot find variable "{param["variable"]}" defined in "{self.attribute_name}" for "{self.path}"'
                        raise DictConsistencyError(msg, 14, self.xmlfiles)
                    param["variable"] = variable
                    if identifier:
                        msg = f'variable "{param["variable"]}" defined in "{self.attribute_name}" for "{self.path}" is a dynamic variable'
                        raise DictConsistencyError(msg, 15, self.xmlfiles)
                else:
                    del param["variable"]
            params[param.pop("key")] = param
        return params


class JinjaCalculation(Calculation):
    attribute_name: Literal[
        "frozen",
        "hidden",
        "mandatory",
        "empty",
        "disabled",
        "default",
        "validators",
        "choices",
        "dynamic",
    ]
    jinja: StrictStr
    params: Optional[List[Param]] = None
    return_type: BASETYPE = None
    description: Optional[StrictStr] = None

    def _jinja_to_function(
        self,
        function,
        return_type,
        multi,
        objectspace,
        *,
        add_help=False,
        params: Optional[dict] = None,
    ):
        variable = objectspace.paths[self.path]
        jinja_path = f"{self.attribute_name}_{self.path}"
        idx = 0
        while jinja_path in objectspace.jinja:
            jinja_path = f"{self.attribute_name}_{self.path}_{idx}"
            idx += 1
        objectspace.jinja[jinja_path] = self.jinja
        default = {
            "function": function,
            "params": {
                "__internal_jinja": jinja_path,
                "__internal_type": return_type,
                "__internal_multi": multi,
                "__internal_files": self.xmlfiles,
                "__internal_attribute": self.attribute_name,
                "__internal_variable": self.path,
            },
        }
        if self.default_values:
            default["params"]["__default_value"] = self.default_values
        if add_help:
            default["help"] = function + "_help"
        if self.params:
            default["params"] |= self.get_params(objectspace)
        if params:
            default["params"] |= params
        if self.ori_path is None:
            path = self.path
        else:
            path = self.ori_path
        for sub_variable, identifier, true_path in get_jinja_variable_to_param(
            path,
            self.jinja,
            objectspace,
            variable.xmlfiles,
            objectspace.functions,
            self.path_prefix,
            self.version,
            self.namespace,
        ):
            if true_path in default["params"]:
                continue
            if isinstance(sub_variable, dict):
                default["params"][true_path] = {
                    "type": "value",
                    "value": sub_variable,
                }
            else:
                default["params"][true_path] = {
                    "type": "variable",
                    "variable": sub_variable,
                }
                if self.version != "1.0":
                    default["params"][true_path]["propertyerror"] = False
                if identifier:
                    default["params"][true_path]["identifier"] = identifier
        return default

    def to_function(
        self,
        objectspace,
    ) -> dict:
        if self.attribute_name == "default":
            if self.return_type:
                raise Exception("return_type not allowed!")
            variable = objectspace.paths[self.path]
            return_type = variable.type
            if self.inside_list:
                multi = False
            elif self.path in objectspace.followers:
                multi = objectspace.multis[self.path] == "submulti"
            else:
                multi = self.path in objectspace.multis
            return self._jinja_to_function(
                "jinja_to_function",
                return_type,
                multi,
                objectspace,
            )
        elif self.attribute_name == "validators":
            if self.return_type:
                raise Exception("pfff")
            return self._jinja_to_function(
                "valid_with_jinja",
                "string",
                False,
                objectspace,
            )
        elif self.attribute_name in PROPERTY_ATTRIBUTE:
            if self.return_type:
                raise Exception("return_type not allowed!")
            return self._jinja_to_function(
                "jinja_to_property",
                "string",
                False,
                objectspace,
                add_help=True,
                params={None: [self.attribute_name], "when": True, "inverse": False},
            )
        elif self.attribute_name == "choices":
            return_type = self.return_type
            if return_type is None:
                return_type = "string"
            return self._jinja_to_function(
                "jinja_to_function",
                return_type,
                not self.inside_list,
                objectspace,
            )
        elif self.attribute_name == "dynamic":
            return self._jinja_to_function(
                "jinja_to_function",
                "string",
                True,
                objectspace,
            )
        raise Exception("hu?")


class _VariableCalculation(Calculation):
    variable: StrictStr
    propertyerror: bool = True
    allow_none: bool = False

    def get_variable(
        self,
        objectspace,
    ) -> "Variable":
        if self.ori_path is None:
            path = self.path
        else:
            path = self.ori_path
        variable, identifier = objectspace.paths.get_with_dynamic(
            self.variable,
            self.path_prefix,
            path,
            self.version,
            self.namespace,
            self.xmlfiles,
        )
        if variable and not isinstance(variable, objectspace.variable):
            # FIXME remove the pfff
            raise Exception("pfff it's a family")
        return variable, identifier

    def get_params(
        self,
        objectspace,
        variable: "Variable",
        identifier: Optional[str],
        *,
        needs_multi: Optional[bool] = None,
    ):
        if not variable:
            msg = f'Variable not found "{self.variable}" for attribut "{self.attribute_name}" for variable "{self.path}"'
            raise DictConsistencyError(msg, 88, self.xmlfiles)
        param = {
            "type": "variable",
            "variable": variable,
            "propertyerror": self.propertyerror,
        }
        if identifier:
            param["identifier"] = identifier
        params = {None: [param]}
        if self.default_values:
            params["__default_value"] = self.default_values
        if self.allow_none:
            params["allow_none"] = True
        if needs_multi is None:
            if self.attribute_name != "default":
                needs_multi = True
            else:
                needs_multi = self.path in objectspace.multis
        calc_variable_is_multi = variable.path in objectspace.multis
        if not calc_variable_is_multi:
            if variable.path in objectspace.paths._dynamics and (
                identifier is None or identifier[-1] is None
            ):
                self_dyn_path = objectspace.paths._dynamics.get(self.path)
                if self_dyn_path is not None:
                    var_dyn_path = objectspace.paths._dynamics[variable.path]
                    if self_dyn_path != var_dyn_path and not self_dyn_path.startswith(
                        f"{var_dyn_path}."
                    ):
                        calc_variable_is_multi = True
                else:
                    calc_variable_is_multi = True
            elif identifier and "{{ identifier }}" in identifier:
                calc_variable_is_multi = True
        if needs_multi:
            if calc_variable_is_multi:
                if self.inside_list:
                    msg = f'the variable "{self.path}" has an invalid attribute "{self.attribute_name}", the variable "{variable.path}" is multi but is inside a list'
                    raise DictConsistencyError(msg, 18, self.xmlfiles)
            elif not self.inside_list:
                msg = f'the variable "{self.path}" has an invalid attribute "{self.attribute_name}", the variable "{variable.path}" is not multi but is not inside a list'
                raise DictConsistencyError(msg, 20, self.xmlfiles)
        elif self.inside_list:
            msg = f'the variable "{self.path}" has an invalid attribute "{self.attribute_name}", it\'s a list'
            raise DictConsistencyError(msg, 23, self.xmlfiles)
        elif calc_variable_is_multi:
            if (
                variable.multi
                or variable.path.rsplit(".", 1)[0] != self.path.rsplit(".", 1)[0]
            ):
                # it's not a follower or not in same leadership
                msg = f'the variable "{self.path}" has an invalid attribute "{self.attribute_name}", the variable "{variable.path}" is a multi'
                raise DictConsistencyError(msg, 21, self.xmlfiles)
            else:
                params[None][0]["index"] = {"index": {"type": "index"}}
        return params


class VariableCalculation(_VariableCalculation):
    attribute_name: Literal["default", "choices", "dynamic"]
    optional: bool = False

    def to_function(
        self,
        objectspace,
    ) -> dict:
        if self.attribute_name != "default" and self.optional:
            msg = f'"{self.attribute_name}" variable shall not have an "optional" attribute for variable "{self.variable}"'
            raise DictConsistencyError(msg, 33, self.xmlfiles)
        variable, identifier = self.get_variable(objectspace)
        if not variable and self.optional:
            raise VariableCalculationDependencyError()
        params = self.get_params(
            objectspace,
            variable,
            identifier,
        )
        return {
            "function": "calc_value",
            "params": params,
        }


class VariablePropertyCalculation(_VariableCalculation):
    attribute_name: Literal[*PROPERTY_ATTRIBUTE]
    when: Any = undefined
    when_not: Any = undefined

    def to_function(
        self,
        objectspace,
    ) -> dict:
        variable, identifier = self.get_variable(objectspace)
        params = self.get_params(
            objectspace,
            variable,
            identifier,
            needs_multi=False,
        )
        variable = params[None][0]["variable"]
        if self.when is not undefined:
            if self.version == "1.0":
                msg = f'when is not allowed in format version 1.0 for attribute "{self.attribute_name}" for variable "{self.path}"'
                raise DictConsistencyError(msg, 103, variable.xmlfiles)
            if self.when_not is not undefined:
                msg = f'the variable "{self.path}" has an invalid attribute "{self.attribute_name}", when and when_not cannot set together'
                raise DictConsistencyError(msg, 31, variable.xmlfiles)
            when = self.when
            inverse = False
        elif self.when_not is not undefined:
            if self.version == "1.0":
                msg = f'when_not is not allowed in format version 1.0 for attribute "{self.attribute_name}" for variable "{self.path}"'
                raise DictConsistencyError(msg, 104, variable.xmlfiles)
            when = self.when_not
            inverse = True
        else:
            if variable.type != "boolean":
                raise Exception("only boolean!")
            when = True
            inverse = False
        params[None].insert(0, self.attribute_name)
        params["when"] = when
        params["inverse"] = inverse
        return {
            "function": "variable_to_property",
            "params": params,
            "help": "variable_to_property",
        }


class InformationCalculation(Calculation):
    attribute_name: Literal["default", "choice", "dynamic"]
    information: StrictStr
    variable: Optional[StrictStr]

    def to_function(
        self,
        objectspace,
    ) -> dict:
        params = {
            None: [
                {
                    "type": "information",
                    "information": self.information,
                }
            ]
        }
        if self.variable:
            if self.ori_path is None:
                path = self.path
            else:
                path = self.ori_path
            variable, identifier = objectspace.paths.get_with_dynamic(
                self.variable,
                self.path_prefix,
                path,
                self.version,
                self.namespace,
                self.xmlfiles,
            )
            if variable is None or identifier is not None:
                raise Exception("pfff")
            params[None][0]["variable"] = variable
        if self.default_values:
            params["__default_value"] = self.default_values
        return {
            "function": "calc_value",
            "params": params,
        }


class _IdentifierCalculation(Calculation):
    identifier: Optional[int] = None

    def get_identifier(self) -> dict:
        identifier = {"type": "identifier"}
        if self.identifier is not None:
            identifier["identifier"] = self.identifier
        return identifier


class IdentifierCalculation(_IdentifierCalculation):
    attribute_name: Literal["default", "choice", "dynamic"]

    def to_function(
        self,
        objectspace,
    ) -> dict:
        identifier = {"type": "identifier"}
        if self.identifier is not None:
            identifier["identifier"] = self.identifier
        return {
            "function": "calc_value",
            "params": {None: [self.get_identifier()]},
        }


class IdentifierPropertyCalculation(_IdentifierCalculation):
    attribute_name: Literal[*PROPERTY_ATTRIBUTE]
    when: Any = undefined
    when_not: Any = undefined

    def to_function(
        self,
        objectspace,
    ) -> dict:
        if self.version == "1.0":
            msg = f'when is not allowed in format version 1.0 for attribute "{self.attribute_name}"'
            raise DictConsistencyError(msg, 105, variable.xmlfiles)
        if self.when is not undefined:
            if self.when_not is not undefined:
                msg = f'the identifier has an invalid attribute "{self.attribute_name}", when and when_not cannot set together'
                raise DictConsistencyError(msg, 35, variable.xmlfiles)
            when = self.when
            inverse = False
        elif self.when_not is not undefined:
            when = self.when_not
            inverse = True
        else:
            msg = f'the identifier has an invalid attribute "{self.attribute_name}", when and when_not cannot set together'
            raise DictConsistencyError
        params = {
            None: [self.attribute_name, self.get_identifier()],
            "when": when,
            "inverse": inverse,
        }
        return {
            "function": "variable_to_property",
            "params": params,
            "help": "variable_to_property",
        }


class IndexCalculation(Calculation):
    attribute_name: Literal["default", "choice", "dynamic"]

    def to_function(
        self,
        objectspace,
    ) -> dict:
        if self.path not in objectspace.followers:
            msg = f'the variable "{self.path}" is not a follower, so cannot have index type for "{self.attribute_name}"'
            raise DictConsistencyError(msg, 60, self.xmlfiles)
        return {
            "function": "calc_value",
            "params": {None: [{"type": "index"}]},
        }


CALCULATION_TYPES = {
    "jinja": JinjaCalculation,
    "variable": VariableCalculation,
    "information": InformationCalculation,
    "identifier": IdentifierCalculation,
    "suffix": IdentifierCalculation,
    "index": IndexCalculation,
}
CALCULATION_PROPERTY_TYPES = {
    "jinja": JinjaCalculation,
    "variable": VariablePropertyCalculation,
    "information": InformationCalculation,
    "identifier": IdentifierPropertyCalculation,
    "index": IndexCalculation,
}
BASETYPE_CALC = Union[StrictBool, StrictInt, StrictFloat, StrictStr, Calculation, None]


class Family(BaseModel):
    name: str
    description: Optional[str] = None
    type: Literal["family", "leadership", "dynamic"] = "family"
    path: str
    help: Optional[str] = None
    mode: Optional[str] = None
    hidden: Union[bool, Calculation] = False
    disabled: Union[bool, Calculation] = False
    namespace: Optional[str]
    version: str
    xmlfiles: List[str] = []

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)


class Dynamic(Family):
    # None only for format 1.0
    variable: str = None
    dynamic: Union[List[Union[StrictStr, Calculation]], Calculation]


class Variable(BaseModel):
    # type will be set dynamically in `annotator/value.py`, default is None
    type: str = None
    name: str
    description: Optional[str] = None
    default: Union[List[BASETYPE_CALC], BASETYPE_CALC] = None
    choices: Optional[Union[List[BASETYPE_CALC], Calculation]] = None
    regexp: Optional[str] = None
    params: Optional[List[Param]] = None
    validators: Optional[List[Calculation]] = None
    multi: Optional[bool] = None
    unique: Optional[bool] = None
    help: Optional[str] = None
    hidden: Union[bool, Calculation] = False
    disabled: Union[bool, Calculation] = False
    mandatory: Union[None, bool, Calculation] = None
    empty: Union[None, bool, Calculation] = True
    auto_save: bool = False
    mode: Optional[str] = None
    test: Optional[list] = None
    examples: Optional[list] = None
    path: str
    namespace: Optional[str]
    version: str
    path_prefix: Optional[str]
    xmlfiles: List[str] = []

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)


class SymLink(BaseModel):
    type: Literal["symlink"] = "symlink"
    name: str
    path: str
    opt: Variable
    namespace: Optional[str]
    version: str
    path_prefix: Optional[str]
    xmlfiles: List[str] = []

    model_config = ConfigDict(extra="forbid")



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_1.rougail.var2'] = "{{ rougail.dyn1.var }}\n"
dict_env['default_2.rougail.var2'] = "{{ rougail.dyn1.var }}\n"
option_3 = IntOption(name="var", doc="a suffix variable", multi=True, default=[1, 2], default_multi=1, properties=frozenset({"mandatory", "standard"}), informations={'type': 'number'})
option_5 = StrOption(name="var", doc="a variable inside dynamic family", default="val", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_4 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_3)))), children=[option_5], properties=frozenset({"standard"}), informations={'dynamic_variable': '1.rougail.var'})
option_6 = StrOption(name="var2", doc="a variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_1.rougail.var2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_0family_dynamic_jinja_number/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("1.rougail.var2"), 'rougail.dyn1.var': ParamDynOption(option_5, ["1"], notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, optiondescription_4, option_6], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_9 = IntOption(name="var", doc="a suffix variable", multi=True, default=[1, 2], default_multi=1, properties=frozenset({"mandatory", "standard"}), informations={'type': 'number'})
option_11 = StrOption(name="var", doc="a variable inside dynamic family", default="val", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_10 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_9)))), children=[option_11], properties=frozenset({"standard"}), informations={'dynamic_variable': '2.rougail.var'})
option_12 = StrOption(name="var2", doc="a variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_2.rougail.var2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_0family_dynamic_jinja_number/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("2.rougail.var2"), 'rougail.dyn1.var': ParamDynOption(option_11, ["1"], notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_8 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_9, optiondescription_10, option_12], properties=frozenset({"standard"}))
optiondescription_7 = OptionDescription(name="2", doc="2", children=[optiondescription_8], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_7])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="mode_conteneur_actif", doc="Description", default="non", properties=frozenset({"mandatory", "normal"}))
option_3 = UsernameOption(name="owner", doc="owner", default="nobody", properties=frozenset({"mandatory", "normal"}))
option_4 = UsernameOption(name="group", doc="group", default="nobody", properties=frozenset({"mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="general", children=[option_2, option_3, option_4], properties=frozenset({"normal"}))
optiondescription_40 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_13 = SymLinkOption(name="group", opt=option_4)
option_14 = FilenameOption(name="name", doc="name", default="/etc/file")
option_15 = SymLinkOption(name="owner", opt=option_3)
option_16 = StrOption(name="source", doc="source", default="file")
option_12 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_11 = OptionDescription(name="file", doc="file", children=[option_13, option_14, option_15, option_16, option_12])
option_19 = SymLinkOption(name="group", opt=option_4)
option_20 = FilenameOption(name="name", doc="name", default="/etc/file2")
option_21 = SymLinkOption(name="owner", opt=option_3)
option_22 = StrOption(name="source", doc="source", default="file2")
option_18 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_17 = OptionDescription(name="file2", doc="file2", children=[option_19, option_20, option_21, option_22, option_18])
optiondescription_17.impl_set_information('engine', "jinja")
optiondescription_10 = OptionDescription(name="files", doc="files", children=[optiondescription_11, optiondescription_17])
option_9 = BoolOption(name="activate", doc="activate", default=True)
option_23 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_42 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_10, option_9, option_23])
optiondescription_42.impl_set_information('type', "service")
optiondescription_41 = OptionDescription(name="services", doc="services", children=[optiondescription_42], properties=frozenset({"hidden"}))
optiondescription_39 = OptionDescription(name="1", doc="1", children=[optiondescription_40, optiondescription_41])
option_6 = StrOption(name="mode_conteneur_actif", doc="Description", default="non", properties=frozenset({"mandatory", "normal"}))
option_7 = UsernameOption(name="owner", doc="owner", default="nobody", properties=frozenset({"mandatory", "normal"}))
option_8 = UsernameOption(name="group", doc="group", default="nobody", properties=frozenset({"mandatory", "normal"}))
optiondescription_5 = OptionDescription(name="general", doc="general", children=[option_6, option_7, option_8], properties=frozenset({"normal"}))
optiondescription_44 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_5], properties=frozenset({"normal"}))
option_28 = SymLinkOption(name="group", opt=option_8)
option_29 = FilenameOption(name="name", doc="name", default="/etc/file")
option_30 = SymLinkOption(name="owner", opt=option_7)
option_31 = StrOption(name="source", doc="source", default="file")
option_27 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_26 = OptionDescription(name="file", doc="file", children=[option_28, option_29, option_30, option_31, option_27])
option_34 = SymLinkOption(name="group", opt=option_8)
option_35 = FilenameOption(name="name", doc="name", default="/etc/file2")
option_36 = SymLinkOption(name="owner", opt=option_7)
option_37 = StrOption(name="source", doc="source", default="file2")
option_33 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_32 = OptionDescription(name="file2", doc="file2", children=[option_34, option_35, option_36, option_37, option_33])
optiondescription_32.impl_set_information('engine', "jinja")
optiondescription_25 = OptionDescription(name="files", doc="files", children=[optiondescription_26, optiondescription_32])
option_24 = BoolOption(name="activate", doc="activate", default=True)
option_38 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_46 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_25, option_24, option_38])
optiondescription_46.impl_set_information('type', "service")
optiondescription_45 = OptionDescription(name="services", doc="services", children=[optiondescription_46], properties=frozenset({"hidden"}))
optiondescription_43 = OptionDescription(name="2", doc="2", children=[optiondescription_44, optiondescription_45])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_39, optiondescription_43])



"""
Config file for Rougail-structural_commandline

Silique (https://www.silique.fr)
Copyright (C) 2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


def get_rougail_config(
    *,
    backward_compatibility=True,
) -> dict:
    options = """
structural_commandline:
  description: Configuration rougail-structural_commandline
  commandline: false
  add_extra_options:
    description: Add extra options to tiramisu-cmdline-parser
    default: true
"""
    return {
        "name": "exporter",
        "process": "structural",
        "options": options,
        "level": 20,
    }


__all__ = "get_rougail_config"



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_2 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_1 = OptionDescription(name="general", doc="general", children=[option_2], properties=frozenset({"standard"}))
option_5 = StrOption(name="leader", doc="leader", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_6 = StrOption(name="follower1", doc="follower1", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_7 = StrOption(name="follower2", doc="follower2", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_4 = Leadership(name="leader", doc="leader", children=[option_5, option_6, option_7], properties=frozenset({"basic"}))
optiondescription_3 = OptionDescription(name="general1", doc="general1", children=[optiondescription_4], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_3])



from shutil import rmtree  #, copyfile, copytree
from os import getcwd, makedirs
from os.path import isfile, join, isdir
from pytest import fixture, raises
from os import listdir
from json import load

from rougail import RougailUpgrade, RougailConfig
from rougail.error import DictConsistencyError

#================================================
class Variable:
    description: str
#================================================
#================================================


dico_dirs = 'tests/dictionaries_old'
dest_dir = 'tmp'


test_ok = set()

for test in listdir(dico_dirs):
    if isdir(join(dico_dirs, test)):
        test_ok.add(test)

excludes = set([])
test_ok -= excludes
#test_ok = ['10load_disabled_if_in_fill']

ORI_DIR = getcwd()

debug = False
#debug = True

test_ok = list(test_ok)
test_ok.sort()
if isdir(dest_dir):
    rmtree(dest_dir)
makedirs(dest_dir)
#copyfile(join(dico_dirs, '__init__.py'), join(dest_dir, '__init__.py'))
#RougailConfig['variable_namespace'] = 'configuration'


@fixture(scope="module", params=test_ok)
def test_dir(request):
    return request.param


def launch_test(test_dir, ext):
    result_dest_dir = join(dico_dirs, test_dir.rsplit('/', 1)[1], 'result')
    new_dest_dir = join(dico_dirs, test_dir.rsplit('/', 1)[1], dest_dir)
    #FIXME
    if isdir(new_dest_dir):
        rmtree(new_dest_dir)
    makedirs(join(new_dest_dir, 'rougail'))
    rougailconfig = RougailConfig.copy()
    rougailconfig['upgrade'] = True
    # rougailconfig_dest = RougailConfig.copy()
    # rougailconfig_dest['functions_files'] = [join(dico_dirs, '../eosfunc/test.py')]
    rougailconfig['main_dictionaries'] = [join(test_dir, 'xml')]
    upgrade_dir = join(new_dest_dir, 'rougail')
    rougailconfig["upgrade_options.main_dictionaries"] = [upgrade_dir]
    # rougailconfig_dest['main_dictionaries'] = [join(new_dest_dir, 'rougail')]
    extra_dictionaries = {}
    if isdir(join(test_dir, ext, 'extra_dirs')):
        extras = listdir(join(test_dir, ext, 'extra_dirs'))
        extras.sort()
        for extra in extras:
            subfolder = join(test_dir, ext, 'extra_dirs', extra)
            if isdir(subfolder):
                extra_dictionaries[extra] = [subfolder]
        makedirs(join(new_dest_dir, 'extra'))
    if extra_dictionaries:
        rougailconfig['extra_dictionaries'] = extra_dictionaries
        rougailconfig["upgrade_options.extra_dictionary"] = join(new_dest_dir, 'extra')
    # rougailconfig_dest['extra_dictionaries'] = extra_dictionaries
    upgrade = RougailUpgrade(rougailconfig=rougailconfig)
#    upgrade.load_dictionaries(join(new_dest_dir, 'rougail'),
#                              extra_dstfolder=new_dest_dir,
#                              services_dstfolder=join(new_dest_dir, 'services'),
#                              )
    upgrade.run()
    ori_lists = set()
    find_files(result_dest_dir,
               [],
               ori_lists,
               )
    new_lists = set()
    find_files(new_dest_dir,
               [],
               new_lists,
               )
    assert ori_lists == new_lists
    for file in ori_lists:
        ori_filename = join(result_dest_dir, file)
        src_filename = join(new_dest_dir, file)
        with open(ori_filename) as fh:
            ori_file = fh.read()
        with open(src_filename) as fh:
            src_file = fh.read()
        assert ori_file == src_file, f'file {ori_filename} and {src_filename} are differents'
    #
    rmtree(new_dest_dir)


def find_files(dirname: str,
               root: list,
               files: set,
               ) -> None:
    for filename in listdir(dirname):
        if filename.startswith('.'):
            continue
        abs_dirname = join(dirname, filename)
        root_file = root + [filename]
        if isdir(join(dirname, filename)):
            find_files(abs_dirname,
                       root_file,
                       files,
                       )
        else:
            files.add(join(*root_file))


def test_dictionary(test_dir):
    assert getcwd() == ORI_DIR
    test_dir = join(dico_dirs, test_dir)
# FIXME   for ext in ['xml', 'yml']:
#        launch_test(test_dir, ext)
    launch_test(test_dir, 'xml')
    assert getcwd() == ORI_DIR



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
class Regexp_option_2(RegexpOption):
    __slots__ = tuple()
    _type = 'value'
Regexp_option_2._regexp = re_compile(r"^#(?:[0-9a-f]{3}){1,2}$")

option_2 = Regexp_option_2(name="var", doc="a first variable", default="#a1a1a1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'regexp', 'test': ('#b1b1b1', '#b2b2b2')})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_4 = BoolOption(name="server_deployed", doc="server_deployed", default=False, properties=frozenset({"hidden", "mandatory", "normal"}))
option_2 = StrOption(name="my_var", doc="my_var", default="no", properties=frozenset({"basic", "force_store_value", "mandatory", Calculation(func.calc_value, Params(ParamValue('frozen'), kwargs={'condition': ParamOption(option_4, notraisepropertyerror=True), 'expected': ParamValue(True)}), func.calc_value_property_help)}))
option_3 = StrOption(name="my_var1", doc="my_var1", default="no", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="général", children=[option_2, option_3, option_4], properties=frozenset({"basic"}))
optiondescription_22 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"basic"}))
option_6 = StrOption(name="description", doc="description", default="Exportation de la base de ejabberd", properties=frozenset({"mandatory", "normal"}))
option_7 = ChoiceOption(name="day", doc="day", values=('none', 'daily', 'weekly', 'monthly'), default="none", properties=frozenset({"mandatory", "normal"}))
option_8 = ChoiceOption(name="mode", doc="mode", values=('pre', 'post'), default="pre", properties=frozenset({"mandatory", "normal"}))
optiondescription_5 = OptionDescription(name="ejabberd", doc="ejabberd", children=[option_6, option_7, option_8], properties=frozenset({"normal"}))
optiondescription_23 = OptionDescription(name="extra", doc="extra", children=[optiondescription_5], properties=frozenset({"normal"}))
option_10 = StrOption(name="description", doc="description", default="test", properties=frozenset({"mandatory", "normal"}))
optiondescription_9 = OptionDescription(name="external", doc="external", children=[option_10], properties=frozenset({"normal", Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_6, notraisepropertyerror=True), 'expected': ParamValue("non")}), func.calc_value_property_help)}))
optiondescription_24 = OptionDescription(name="extra1", doc="extra1", children=[optiondescription_9], properties=frozenset({"normal"}))
optiondescription_21 = OptionDescription(name="1", doc="1", children=[optiondescription_22, optiondescription_23, optiondescription_24])
option_14 = BoolOption(name="server_deployed", doc="server_deployed", default=False, properties=frozenset({"hidden", "mandatory", "normal"}))
option_12 = StrOption(name="my_var", doc="my_var", default="no", properties=frozenset({"basic", "force_store_value", "mandatory", Calculation(func.calc_value, Params(ParamValue('frozen'), kwargs={'condition': ParamOption(option_14, notraisepropertyerror=True), 'expected': ParamValue(True)}), func.calc_value_property_help)}))
option_13 = StrOption(name="my_var1", doc="my_var1", default="no", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_11 = OptionDescription(name="general", doc="général", children=[option_12, option_13, option_14], properties=frozenset({"basic"}))
optiondescription_26 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_11], properties=frozenset({"basic"}))
option_16 = StrOption(name="description", doc="description", default="Exportation de la base de ejabberd", properties=frozenset({"mandatory", "normal"}))
option_17 = ChoiceOption(name="day", doc="day", values=('none', 'daily', 'weekly', 'monthly'), default="none", properties=frozenset({"mandatory", "normal"}))
option_18 = ChoiceOption(name="mode", doc="mode", values=('pre', 'post'), default="pre", properties=frozenset({"mandatory", "normal"}))
optiondescription_15 = OptionDescription(name="ejabberd", doc="ejabberd", children=[option_16, option_17, option_18], properties=frozenset({"normal"}))
optiondescription_27 = OptionDescription(name="extra", doc="extra", children=[optiondescription_15], properties=frozenset({"normal"}))
option_20 = StrOption(name="description", doc="description", default="test", properties=frozenset({"mandatory", "normal"}))
optiondescription_19 = OptionDescription(name="external", doc="external", children=[option_20], properties=frozenset({"normal", Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_16, notraisepropertyerror=True), 'expected': ParamValue("non")}), func.calc_value_property_help)}))
optiondescription_28 = OptionDescription(name="extra1", doc="extra1", children=[optiondescription_19], properties=frozenset({"normal"}))
optiondescription_25 = OptionDescription(name="2", doc="2", children=[optiondescription_26, optiondescription_27, optiondescription_28])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_21, optiondescription_25])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_1 = FilenameOption(name="file_name", doc="file_name", multi=True, default=['/etc/mailname', '/etc/mailname2'], default_multi="/etc/mailname", properties=frozenset({"mandatory", "normal"}))
option_2 = FilenameOption(name="file_name2", doc="file_name2", multi=True, default=['/etc/mailname3', '/etc/mailname4'], default_multi="/etc/mailname3", properties=frozenset({"mandatory", "normal"}))
option_3 = StrOption(name="var", doc="var", multi=True, default=['mailname', 'mailname2'], default_multi="mailname", properties=frozenset({"mandatory", "normal"}))
optiondescription_34 = OptionDescription(name="rougail", doc="Rougail", children=[option_1, option_2, option_3], properties=frozenset({"normal"}))
option_11 = SymLinkOption(name="name", opt=option_1)
option_12 = StrOption(name="source", doc="source", default="mailname")
option_13 = SymLinkOption(name="variable", opt=option_3)
option_10 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_9 = OptionDescription(name="mailname", doc="mailname", children=[option_11, option_12, option_13, option_10])
option_16 = SymLinkOption(name="name", opt=option_2)
option_17 = StrOption(name="source", doc="source", default="mailname2")
option_18 = SymLinkOption(name="variable", opt=option_3)
option_15 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_14 = OptionDescription(name="mailname2", doc="mailname2", children=[option_16, option_17, option_18, option_15])
optiondescription_14.impl_set_information('engine', "jinja")
optiondescription_8 = OptionDescription(name="files", doc="files", children=[optiondescription_9, optiondescription_14])
option_7 = BoolOption(name="activate", doc="activate", default=True)
option_19 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_36 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_8, option_7, option_19])
optiondescription_36.impl_set_information('type', "service")
optiondescription_35 = OptionDescription(name="services", doc="services", children=[optiondescription_36], properties=frozenset({"hidden"}))
optiondescription_33 = OptionDescription(name="1", doc="1", children=[optiondescription_34, optiondescription_35])
option_4 = FilenameOption(name="file_name", doc="file_name", multi=True, default=['/etc/mailname', '/etc/mailname2'], default_multi="/etc/mailname", properties=frozenset({"mandatory", "normal"}))
option_5 = FilenameOption(name="file_name2", doc="file_name2", multi=True, default=['/etc/mailname3', '/etc/mailname4'], default_multi="/etc/mailname3", properties=frozenset({"mandatory", "normal"}))
option_6 = StrOption(name="var", doc="var", multi=True, default=['mailname', 'mailname2'], default_multi="mailname", properties=frozenset({"mandatory", "normal"}))
optiondescription_38 = OptionDescription(name="rougail", doc="Rougail", children=[option_4, option_5, option_6], properties=frozenset({"normal"}))
option_24 = SymLinkOption(name="name", opt=option_4)
option_25 = StrOption(name="source", doc="source", default="mailname")
option_26 = SymLinkOption(name="variable", opt=option_6)
option_23 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_22 = OptionDescription(name="mailname", doc="mailname", children=[option_24, option_25, option_26, option_23])
option_29 = SymLinkOption(name="name", opt=option_5)
option_30 = StrOption(name="source", doc="source", default="mailname2")
option_31 = SymLinkOption(name="variable", opt=option_6)
option_28 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_27 = OptionDescription(name="mailname2", doc="mailname2", children=[option_29, option_30, option_31, option_28])
optiondescription_27.impl_set_information('engine', "jinja")
optiondescription_21 = OptionDescription(name="files", doc="files", children=[optiondescription_22, optiondescription_27])
option_20 = BoolOption(name="activate", doc="activate", default=True)
option_32 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_40 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_21, option_20, option_32])
optiondescription_40.impl_set_information('type', "service")
optiondescription_39 = OptionDescription(name="services", doc="services", children=[optiondescription_40], properties=frozenset({"hidden"}))
optiondescription_37 = OptionDescription(name="2", doc="2", children=[optiondescription_38, optiondescription_39])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_33, optiondescription_37])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['dynamic_1.rougail.dyn{{ identifier }}'] = "{% for val in _.var %}\n{{ loop.index }}\n{% endfor %}\n"
dict_env['dynamic_2.rougail.dyn{{ identifier }}'] = "{% for val in _.var %}\n{{ loop.index }}\n{% endfor %}\n"
option_3 = StrOption(name="var", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_5 = StrOption(name="var", doc="a dynamic variable", default="val", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_4 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="A dynamic family", identifiers=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("dynamic_1.rougail.dyn{{ identifier }}"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/60_1family_dynamic_jinja/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("dynamic"), '__internal_variable': ParamValue("1.rougail.dyn{{ identifier }}"), '_.var': ParamOption(option_3, notraisepropertyerror=True)})), children=[option_5], properties=frozenset({"standard"}))
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, optiondescription_4], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_8 = StrOption(name="var", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_10 = StrOption(name="var", doc="a dynamic variable", default="val", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_9 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="A dynamic family", identifiers=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("dynamic_2.rougail.dyn{{ identifier }}"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/60_1family_dynamic_jinja/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("dynamic"), '__internal_variable': ParamValue("2.rougail.dyn{{ identifier }}"), '_.var': ParamOption(option_8, notraisepropertyerror=True)})), children=[option_10], properties=frozenset({"standard"}))
optiondescription_7 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_8, optiondescription_9], properties=frozenset({"standard"}))
optiondescription_6 = OptionDescription(name="2", doc="2", children=[optiondescription_7], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_6])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_rougail.variable'] = "{{test_information }}\n"
option_2 = StrOption(name="variable", doc="a variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_rougail.variable"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/00_8calculation_information/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("rougail.variable"), 'test_information': ParamInformation("test_information", None)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



"""Annotate variable

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from rougail.i18n import _
from rougail.error import DictConsistencyError
from rougail.object_model import Calculation, VariableCalculation
from tiramisu.error import display_list


class Walk:
    """Walk to objectspace to find variable or family"""

    objectspace = None

    def get_variables(self):
        """Iter all variables from the objectspace"""
        for path in self.objectspace.variables:
            yield self.objectspace.paths[path]

    #        yield from get_variables(self.objectspace)

    def get_families(self):
        """Iter all families from the objectspace"""
        for path in self.objectspace.families:
            yield self.objectspace.paths[path]


class Annotator(Walk):  # pylint: disable=R0903
    """Annotate variable"""

    level = 30

    def __init__(
        self,
        objectspace,
        *args,
    ):
        if not objectspace.paths:
            return
        self.objectspace = objectspace
        if self.objectspace.main_namespace:
            self.forbidden_name = [self.objectspace.main_namespace]
            for extra in self.objectspace.extra_dictionaries:
                self.forbidden_name.append(extra)
        else:
            self.forbidden_name = []
        # default type inference from a default value with :term:`basic types`
        self.basic_types = {
            str: "string",
            int: "number",
            bool: "boolean",
            float: "float",
        }
        self.verify_choices()
        self.convert_variable()
        self.convert_test()
        self.convert_examples()
        self.convert_help()

    def convert_variable(self):
        """convert variable"""
        for variable in self.get_variables():
            if variable.version != "1.0":
                if variable.type == "symlink":
                    continue
                self._convert_variable_inference(variable)
        for variable in self.get_variables():
            if variable.type == "symlink":
                continue
            if variable.version != "1.0":
                self._default_variable_copy_informations(variable)
            if variable.multi is None:
                variable.multi = False
            if variable.type is None:
                variable.type = "string"
            self.objectspace.informations.add(variable.path, "type", variable.type)
            self._convert_variable(variable)

    def _convert_variable_inference(
        self,
        variable,
    ) -> None:
        # variable has no type
        if variable.type is None:
            # choice type inference from the `choices` attribute
            if variable.choices is not None:
                variable.type = "choice"
            elif variable.regexp is not None:
                variable.type = "regexp"
            elif variable.default not in [None, []]:
                if isinstance(variable.default, list):
                    tested_value = variable.default[0]
                else:
                    tested_value = variable.default
                variable.type = self.basic_types.get(type(tested_value), None)
        # variable has no multi attribute
        if variable.multi is None and not (
            variable.type is None and isinstance(variable.default, VariableCalculation)
        ):
            if variable.path in self.objectspace.leaders:
                variable.multi = True
            else:
                variable.multi = isinstance(variable.default, list)

    def _default_variable_copy_informations(
        self,
        variable,
    ) -> None:
        # if a variable has a variable as default value, that means the type/params or multi should has same value
        if variable.type is not None or not isinstance(
            variable.default, VariableCalculation
        ):
            return
        # copy type and params
        calculated_variable_path = variable.default.variable
        calculated_variable, identifier = self.objectspace.paths.get_with_dynamic(
            calculated_variable_path,
            variable.default.path_prefix,
            variable.path,
            variable.version,
            variable.namespace,
            variable.xmlfiles,
        )
        if calculated_variable is None:
            return
        variable.type = calculated_variable.type
        if variable.params is None and calculated_variable.params is not None:
            variable.params = calculated_variable.params
        # copy multi attribut
        if variable.multi is None:
            calculated_path = calculated_variable.path
            if (
                calculated_path in self.objectspace.leaders
                and variable.path in self.objectspace.followers
                and calculated_path.rsplit(".")[0] == variable.path.rsplit(".")[0]
            ):
                variable.multi = False
            else:
                variable.multi = calculated_variable.multi

    def _convert_variable(
        self,
        variable: dict,
    ) -> None:
        # variable without description: description is the name
        if not variable.description:
            variable.description = variable.name
        if variable.path in self.objectspace.followers:
            if not variable.multi:
                self.objectspace.multis[variable.path] = True
            else:
                self.objectspace.multis[variable.path] = "submulti"
        elif variable.multi:
            self.objectspace.multis[variable.path] = True
        if variable.path in self.objectspace.leaders:
            if not self.objectspace.multis.get(variable.path, False):
                variable.multi = self.objectspace.multis[variable.path] = True
            family = self.objectspace.paths[variable.path.rsplit(".", 1)[0]]
            if variable.hidden:
                family.hidden = variable.hidden
            #            elif family.hidden:
            #                variable.hidden = family.hidden
            variable.hidden = None
        if variable.regexp is not None and variable.type != "regexp":
            msg = _(
                'the variable "{0}" has regexp attribut but has not the "regexp" type'
            ).format(variable.path)
            raise DictConsistencyError(msg, 37, variable.xmlfiles)
        if variable.mandatory is None:
            variable.mandatory = True

    def convert_test(self):
        """Convert variable tests value"""
        for variable in self.get_variables():
            if variable.type == "symlink":
                continue
            if variable.test is None:
                continue
            self.objectspace.informations.add(
                variable.path, "test", tuple(variable.test)
            )

    def convert_examples(self):
        """Convert variable tests value"""
        for variable in self.get_variables():
            if variable.type == "symlink":
                continue
            if variable.examples is None:
                continue
            self.objectspace.informations.add(
                variable.path, "examples", tuple(variable.examples)
            )

    def convert_help(self):
        """Convert variable help"""
        for variable in self.get_variables():
            if not hasattr(variable, "help") or not variable.help:
                continue
            self.objectspace.informations.add(variable.path, "help", variable.help)
            del variable.help

    def verify_choices(self):
        for variable in self.get_variables():
            if variable.type is None and variable.choices:
                # choice type inference from the `choices` attribute
                variable.type = "choice"
            if variable.choices is not None and variable.type != "choice":
                msg = _(
                    'the variable "{0}" has choices attribut but has not the "choice" type'
                ).format(variable.path)
                raise DictConsistencyError(msg, 11, variable.xmlfiles)
            if variable.type != "choice":
                continue
            if variable.default is None:
                continue
            if None in variable.choices and variable.mandatory is None:
                variable.mandatory = False
            if not isinstance(variable.choices, list):
                continue
            choices = variable.choices
            has_calculation = False
            for choice in choices:
                if isinstance(choice, Calculation):
                    has_calculation = True
                    break
            if has_calculation:
                continue

            default = variable.default
            if not isinstance(default, list):
                default = [default]
            for value in default:
                if isinstance(value, Calculation):
                    continue
                if value not in choices:
                    msg = _(
                        'the variable "{0}" has an unvalid default value "{1}" should be in {2}'
                    ).format(
                        variable.path,
                        value,
                        display_list(choices, separator="or", add_quote=True),
                    )
                    raise DictConsistencyError(msg, 26, variable.xmlfiles)



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['frozen_1.rougail.dyn{{ identifier }}.var'] = "{% if suffix == 'val2' %}\ndisabled\n{% endif %}\n"
dict_env['frozen_1.rougail.dyn{{ identifier }}.family.var'] = "{% if suffix == 'val2' %}\ndisabled\n{% endif %}\n"
dict_env['hidden_1.rougail.dyn{{ identifier }}'] = "{% if suffix == 'val2' %}\ndisabled\n{% endif %}\n"
dict_env['frozen_2.rougail.dyn{{ identifier }}.var'] = "{% if suffix == 'val2' %}\ndisabled\n{% endif %}\n"
dict_env['frozen_2.rougail.dyn{{ identifier }}.family.var'] = "{% if suffix == 'val2' %}\ndisabled\n{% endif %}\n"
dict_env['hidden_2.rougail.dyn{{ identifier }}'] = "{% if suffix == 'val2' %}\ndisabled\n{% endif %}\n"
option_4 = StrOption(name="var", doc="a variable", properties=frozenset({"force_default_on_freeze", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("frozen")), kwargs={'__internal_jinja': ParamValue("frozen_1.rougail.dyn{{ identifier }}.var"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_hidden_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("frozen"), '__internal_variable': ParamValue("1.rougail.dyn{{ identifier }}.var"), 'suffix': ParamIdentifier(), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
option_6 = StrOption(name="var", doc="a new variable", properties=frozenset({"force_default_on_freeze", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("frozen")), kwargs={'__internal_jinja': ParamValue("frozen_1.rougail.dyn{{ identifier }}.family.var"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_hidden_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("frozen"), '__internal_variable': ParamValue("1.rougail.dyn{{ identifier }}.family.var"), 'suffix': ParamIdentifier(), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_5 = OptionDescription(name="family", doc="a family", children=[option_6], properties=frozenset({"standard"}))
optiondescription_3 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="a dynamic family", identifiers=["val1", "val2"], children=[option_4, optiondescription_5], properties=frozenset({"standard", Calculation(func['jinja_to_property'], Params((ParamValue("hidden")), kwargs={'__internal_jinja': ParamValue("hidden_1.rougail.dyn{{ identifier }}"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_hidden_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("hidden"), '__internal_variable': ParamValue("1.rougail.dyn{{ identifier }}"), 'suffix': ParamIdentifier(), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}))
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_3], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_10 = StrOption(name="var", doc="a variable", properties=frozenset({"force_default_on_freeze", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("frozen")), kwargs={'__internal_jinja': ParamValue("frozen_2.rougail.dyn{{ identifier }}.var"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_hidden_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("frozen"), '__internal_variable': ParamValue("2.rougail.dyn{{ identifier }}.var"), 'suffix': ParamIdentifier(), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
option_12 = StrOption(name="var", doc="a new variable", properties=frozenset({"force_default_on_freeze", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("frozen")), kwargs={'__internal_jinja': ParamValue("frozen_2.rougail.dyn{{ identifier }}.family.var"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_hidden_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("frozen"), '__internal_variable': ParamValue("2.rougail.dyn{{ identifier }}.family.var"), 'suffix': ParamIdentifier(), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_11 = OptionDescription(name="family", doc="a family", children=[option_12], properties=frozenset({"standard"}))
optiondescription_9 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="a dynamic family", identifiers=["val1", "val2"], children=[option_10, optiondescription_11], properties=frozenset({"standard", Calculation(func['jinja_to_property'], Params((ParamValue("hidden")), kwargs={'__internal_jinja': ParamValue("hidden_2.rougail.dyn{{ identifier }}"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_hidden_suffix/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("hidden"), '__internal_variable': ParamValue("2.rougail.dyn{{ identifier }}"), 'suffix': ParamIdentifier(), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}))
optiondescription_8 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_9], properties=frozenset({"standard"}))
optiondescription_7 = OptionDescription(name="2", doc="2", children=[optiondescription_8], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_7])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_3 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="type", doc="a type family", children=[option_3], properties=frozenset({"standard"}))
option_5 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_4 = OptionDescription(name="description", doc="This is a other great family", children=[option_5], properties=frozenset({"standard"}))
option_7 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_6 = OptionDescription(name="help", doc="a help family", children=[option_7], properties=frozenset({"standard"}), informations={'help': 'This is a other great family'})
option_9 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"advanced", "force_default_on_freeze", "frozen"}), informations={'type': 'string'})
optiondescription_8 = OptionDescription(name="mode", doc="a mode family", children=[option_9], properties=frozenset({"advanced"}))
option_11 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_10 = OptionDescription(name="hidden", doc="an hidden family", children=[option_11], properties=frozenset({"hidden", "standard"}))
option_13 = StrOption(name="my_variable", doc="my_variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_12 = OptionDescription(name="disabled", doc="an disabled family", children=[option_13], properties=frozenset({"disabled", "standard"}))
optiondescription_1 = OptionDescription(name="my_family", doc="This is a great family", children=[optiondescription_2, optiondescription_4, optiondescription_6, optiondescription_8, optiondescription_10, optiondescription_12], properties=frozenset({"disabled", "hidden", "standard"}), informations={'help': 'This is a great family'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_4 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_3 = OptionDescription(name="general", doc="general", children=[option_4], properties=frozenset({"standard"}))
option_7 = StrOption(name="leader", doc="leader", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_8 = StrOption(name="follower1", doc="follower1", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_9 = StrOption(name="follower2", doc="follower2", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_6 = Leadership(name="leader", doc="leader", children=[option_7, option_8, option_9], properties=frozenset({"basic"}))
optiondescription_5 = OptionDescription(name="general1", doc="general1", children=[optiondescription_6], properties=frozenset({"basic"}))
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_3, optiondescription_5], properties=frozenset({"basic"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"basic"}))
option_13 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_12 = OptionDescription(name="general", doc="general", children=[option_13], properties=frozenset({"standard"}))
option_16 = StrOption(name="leader", doc="leader", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_17 = StrOption(name="follower1", doc="follower1", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_18 = StrOption(name="follower2", doc="follower2", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_15 = Leadership(name="leader", doc="leader", children=[option_16, option_17, option_18], properties=frozenset({"basic"}))
optiondescription_14 = OptionDescription(name="general1", doc="general1", children=[optiondescription_15], properties=frozenset({"basic"}))
optiondescription_11 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_12, optiondescription_14], properties=frozenset({"basic"}))
optiondescription_10 = OptionDescription(name="2", doc="2", children=[optiondescription_11], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_10])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="mode_conteneur_actif", doc="No change", default="oui", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="général", children=[option_2], properties=frozenset({"normal"}))
optiondescription_10 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_5 = BoolOption(name="activate", doc="activate", default=True)
option_6 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_12 = OptionDescription(name="testsrv_service", doc="testsrv.service", children=[option_5, option_6])
optiondescription_12.impl_set_information('target', "test")
optiondescription_12.impl_set_information('engine', "none")
optiondescription_12.impl_set_information('type', "service")
optiondescription_11 = OptionDescription(name="services", doc="services", children=[optiondescription_12], properties=frozenset({"hidden"}))
optiondescription_9 = OptionDescription(name="1", doc="1", children=[optiondescription_10, optiondescription_11])
option_4 = StrOption(name="mode_conteneur_actif", doc="No change", default="oui", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_3 = OptionDescription(name="general", doc="général", children=[option_4], properties=frozenset({"normal"}))
optiondescription_14 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_3], properties=frozenset({"normal"}))
option_7 = BoolOption(name="activate", doc="activate", default=True)
option_8 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_16 = OptionDescription(name="testsrv_service", doc="testsrv.service", children=[option_7, option_8])
optiondescription_16.impl_set_information('target', "test")
optiondescription_16.impl_set_information('engine', "none")
optiondescription_16.impl_set_information('type', "service")
optiondescription_15 = OptionDescription(name="services", doc="services", children=[optiondescription_16], properties=frozenset({"hidden"}))
optiondescription_13 = OptionDescription(name="2", doc="2", children=[optiondescription_14, optiondescription_15])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_9, optiondescription_13])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_1 = StrOption(name="var", doc="A suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_3 = StrOption(name="vardyn", doc="A dynamic variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_2 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="A dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_1, notraisepropertyerror=True)), kwargs={'allow_none': ParamValue(True)})), children=[option_3], properties=frozenset({"basic"}), informations={'dynamic_variable': 'var'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, optiondescription_2])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_2 = StrOption(name="var", doc="a variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_4 = StrOption(name="var", doc="a calculated variable", default=Calculation(func['calc_value'], Params((ParamOption(option_2)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_3 = OptionDescription(name="fam1", doc="a family", children=[option_4], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, optiondescription_3], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries_old/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_3 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
option_4 = StrOption(name="activer_ejabberd", doc="No change", default="non", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_2 = OptionDescription(name="general", doc="général", children=[option_3, option_4], properties=frozenset({"normal"}))
optiondescription_1 = OptionDescription(name="rougail", doc="rougail", children=[optiondescription_2], properties=frozenset({"normal"}))
option_8 = StrOption(name="description", doc="description", multi=True, default=['test'], properties=frozenset({"mandatory", "normal"}))
option_9 = StrOption(name="mode", doc="mode", multi=True, default="pre", properties=frozenset({"mandatory", "normal"}))
optiondescription_7 = Leadership(name="leadership", doc="description", children=[option_8, option_9], properties=frozenset({"normal"}))
optiondescription_6 = OptionDescription(name="ejabberd", doc="ejabberd", children=[optiondescription_7], properties=frozenset({"normal"}))
optiondescription_5 = OptionDescription(name="extra", doc="extra", children=[optiondescription_6], properties=frozenset({"normal"}))
option_14 = FilenameOption(name="name", doc="name", default="/etc/mailname")
option_15 = FilenameOption(name="source", doc="source", default="/etc/mailname")
option_16 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_13 = OptionDescription(name="mailname", doc="mailname", children=[option_14, option_15, option_16])
option_18 = FilenameOption(name="name", doc="name", default="/etc/mailname2")
option_19 = FilenameOption(name="source", doc="source", default="/etc/mailname2")
option_20 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_17 = OptionDescription(name="mailname2", doc="mailname2", children=[option_18, option_19, option_20])
optiondescription_17.impl_set_information('engine', "jinja")
optiondescription_12 = OptionDescription(name="files", doc="files", children=[optiondescription_13, optiondescription_17])
option_21 = BoolOption(name="activate", doc="activate", default=True)
option_22 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_11 = OptionDescription(name="test_service", doc="test_service", children=[optiondescription_12, option_21, option_22])
optiondescription_11.impl_set_information('type', "service")
optiondescription_10 = OptionDescription(name="services", doc="services", children=[optiondescription_11], properties=frozenset({"hidden", "normal"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_5, optiondescription_10])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_3 = StrOption(name="leader", doc="the leader", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_4 = StrOption(name="follower1", doc="the follower1", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_5 = StrOption(name="follower2", doc="the follower2", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_6 = StrOption(name="follower3", doc="the follower3", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_2 = Leadership(name="leader", doc="a leadership", children=[option_3, option_4, option_5, option_6], properties=frozenset({"basic"}))
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_2], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries_old/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
from jinja2 import StrictUndefined, DictLoader
from jinja2.sandbox import SandboxedEnvironment
from rougail.annotator.variable import CONVERT_OPTION
def jinja_to_function(__internal_jinja, __internal_type, __internal_multi, **kwargs):
    kw = {}
    for key, value in kwargs.items():
        if '.' in key:
            c_kw = kw
            path, var = key.rsplit('.', 1)
            for subkey in path.split('.'):
                c_kw = c_kw.setdefault(subkey, {})
            c_kw[var] = value
        else:
            kw[key] = value
    values = ENV.get_template(__internal_jinja).render(kw)
    convert = CONVERT_OPTION[__internal_type].get('func', str)
    if __internal_multi:
        return [convert(val) for val in values.split(',')]
    return convert(values)
def valid_with_jinja(value, **kwargs):
    kwargs[kwargs.pop('__internal_key')] = value
    value = jinja_to_function(__internal_type='string', __internal_multi=False, **kwargs)
    if value:
        raise ValueError(value)
func.jinja_to_function = jinja_to_function
func.valid_with_jinja = valid_with_jinja
dict_env = {}
dict_env['rougail.dyn.vardyn'] = "{{ __suffix | calc_val }}"
ENV = SandboxedEnvironment(loader=DictLoader(dict_env), undefined=StrictUndefined)
ENV.compile_templates('jinja_caches', zip=None)
option_3 = StrOption(name="varname", doc="No change", multi=True, default=['val1', 'val2'], default_multi="val1", properties=frozenset({"mandatory", "normal"}))
optiondescription_2 = OptionDescription(name="general", doc="general", children=[option_3], properties=frozenset({"normal"}))
option_5 = StrOption(name="vardyn", doc="No change", default=Calculation(func.jinja_to_function, Params((), kwargs={'__internal_jinja': ParamValue(rougail.dyn.vardyn), '__internal_type': ParamValue(string), '__internal_multi': ParamValue(False), '__suffix': ParamSuffix()})), properties=frozenset({"normal"}))
optiondescription_4 = ConvertDynOptionDescription(name="dyn", doc="dyn", suffixes=Calculation(func.calc_value, Params((ParamOption(option_3, notraisepropertyerror=True)))), children=[option_5], properties=frozenset({"normal"}))
option_7 = StrOption(name="newvar", doc="No change", properties=frozenset({"normal"}))
optiondescription_6 = OptionDescription(name="new", doc="new", children=[option_7], properties=frozenset({"normal"}))
optiondescription_1 = OptionDescription(name="rougail", doc="rougail", children=[optiondescription_2, optiondescription_4, optiondescription_6], properties=frozenset({"normal"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



"""Annotate properties

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Union
from rougail.i18n import _
from rougail.error import DictConsistencyError
from rougail.annotator.variable import Walk
from rougail.object_model import Calculation


PROPERTIES = (
    "hidden",
    "frozen",
    "force_default_on_freeze",
    "force_store_value",
    "disabled",
    "mandatory",
)


class Annotator(Walk):
    """Annotate properties"""

    level = 90

    def __init__(self, objectspace, *args) -> None:
        self.objectspace = objectspace
        self.frozen = {}
        if self.objectspace.paths:
            self.convert_family()
            self.convert_variable()

    def convert_family(self) -> None:
        """convert families"""
        for family in self.get_families():
            self._convert_property(family)
            # collect for force_default_on_freeze
            if family.hidden:
                self.set_variable_frozen(
                    family.path,
                    family.hidden,
                )

    def set_variable_frozen(
        self,
        family_path: str,
        hidden: Union[bool, Calculation],
    ) -> None:
        for variable_path in self.objectspace.parents[family_path]:
            if variable_path in self.objectspace.families:
                # it's a family
                self.set_variable_frozen(
                    variable_path,
                    hidden,
                )
            else:
                # it's a variable
                variable = self.objectspace.paths[variable_path]
                # if frozen is already true or hidden for variable is true => always frozen
                if (
                    self.frozen.get(variable.path) is True
                    or variable.hidden is True
                    or hidden is True
                ):
                    self.frozen[variable.path] = True
                elif variable.path in self.frozen:
                    self.frozen[variable.path].append(hidden)
                else:
                    self.frozen[variable.path] = [hidden]

    def convert_variable(self) -> None:
        """convert variables"""
        for variable in self.get_variables():
            if variable.path.startswith("services."):
                continue
            if variable.type == "symlink":
                continue
            self._convert_variable_property(variable)

    def _convert_variable_property(
        self,
        variable: dict,
    ) -> None:
        """convert properties"""
        path = variable.path
        self._convert_property(variable)
        if variable.hidden:
            if variable.hidden is True:
                self.frozen[path] = True
            elif self.frozen.get(path) is not True:
                self.frozen.setdefault(path, []).append(variable.hidden)
        if path in self.frozen:
            frozen = self.frozen[path]
            if frozen is True:
                value = True
            else:
                value = []
                for calculation in frozen:
                    calculation_copy = calculation.copy()
                    calculation_copy.attribute_name = "frozen"
                    calculation_copy.ori_path = calculation_copy.path
                    calculation_copy.path = path
                    value.append(calculation_copy)
                if len(value) == 1:
                    value = value[0]
            self.objectspace.properties.add(path, "frozen", value)
            if not variable.auto_save:
                # if auto_save, save calculated value
                self.objectspace.properties.add(path, "force_default_on_freeze", True)
        if not variable.empty and self.objectspace.multis.get(variable.path, False):
            # a multi could not have "None" has value
            # to permit it, just add empty="false"
            self.objectspace.properties.add(path, "notempty", True)
        if variable.unique:
            self.objectspace.properties.add(path, "unique", True)
        if variable.unique is False:
            self.objectspace.properties.add(path, "notunique", True)
        if variable.auto_save:
            self.objectspace.properties.add(path, "force_store_value", True)

    def _convert_property(
        self,
        obj: dict,
    ) -> None:
        for prop in PROPERTIES:
            if not hasattr(obj, prop):
                continue
            value = getattr(obj, prop)
            if not value:
                continue
            self.objectspace.properties.add(obj.path, prop, value)
        if obj.mode:
            self.objectspace.properties.add(obj.path, obj.mode, True)



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_3 = StrOption(name="varname", doc="No change", multi=True, default=["a"], default_multi="a", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="general", doc="général", children=[option_3], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_2], properties=frozenset({"standard"}))
option_5 = StrOption(name="var", doc="a varaible", multi=True, default=["a"], default_multi="a", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_7 = StrOption(name="var", doc="var", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_6 = ConvertDynOptionDescription(name="dyn_{{ identifier }}", doc="dyn_{{ identifier }}", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_5)))), children=[option_7], properties=frozenset({"basic"}), informations={'dynamic_variable': 'extra.var'})
optiondescription_4 = OptionDescription(name="extra", doc="extra", group_type=groups.namespace, children=[option_5, optiondescription_6], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_4])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_3 = BoolOption(name="var1", doc="the first variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_4 = BoolOption(name="var2", doc="the second variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_5 = BoolOption(name="var3", doc="the third variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_6 = BoolOption(name="var4", doc="the forth variable", multi=True, default=[False], default_multi=False, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_7 = BoolOption(name="var5", doc="the fifth variable", multi=True, default=[False], default_multi=False, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_8 = BoolOption(name="var6", doc="the sixth variable", multi=True, default=[False], default_multi=False, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_9 = BoolOption(name="var7", doc="the seventh variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_10 = BoolOption(name="var8", doc="the eighth variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, option_4, option_5, option_6, option_7, option_8, option_9, option_10], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"standard"}))
option_13 = BoolOption(name="var1", doc="the first variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_14 = BoolOption(name="var2", doc="the second variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_15 = BoolOption(name="var3", doc="the third variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_16 = BoolOption(name="var4", doc="the forth variable", multi=True, default=[False], default_multi=False, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_17 = BoolOption(name="var5", doc="the fifth variable", multi=True, default=[False], default_multi=False, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_18 = BoolOption(name="var6", doc="the sixth variable", multi=True, default=[False], default_multi=False, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_19 = BoolOption(name="var7", doc="the seventh variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
option_20 = BoolOption(name="var8", doc="the eighth variable", multi=True, default=[True], default_multi=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean'})
optiondescription_12 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_13, option_14, option_15, option_16, option_17, option_18, option_19, option_20], properties=frozenset({"standard"}))
optiondescription_11 = OptionDescription(name="2", doc="2", children=[optiondescription_12], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_11])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['disabled_rougail.leader.follower'] = "{% if _.leader == \"a\" %}\nthe value of \"leader\" is \"a\"\n{% endif %}\n"
option_3 = StrOption(name="leader", doc="a leader", multi=True, default=["a", "b"], properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_4 = StrOption(name="follower", doc="a follower", multi=True, default=Calculation(func['calc_value'], Params((ParamOption(option_3)))), properties=frozenset({"mandatory", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_rougail.leader.follower"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/44_9calculated_default_leadership_leader/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("rougail.leader.follower"), 'when': ParamValue(True), 'inverse': ParamValue(False), '_.leader': ParamOption(option_3, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_2 = Leadership(name="leader", doc="leader", children=[option_3, option_4], properties=frozenset({"standard"}))
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_2], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['choices_var'] = "{% for n in trange(0, 10) %}\n{{ n }}\n{% endfor %}\n"
option_1 = ChoiceOption(name="var", doc="a variable", values=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("choices_var"), '__internal_type': ParamValue("number"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/00_6choice_calculation/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("choices"), '__internal_variable': ParamValue("var")})), default=9, properties=frozenset({"mandatory", "standard"}), informations={'type': 'choice'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1])



"""Rougail method

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from tiramisu import Config, undefined
from tiramisu.error import PropertiesOptionError, LeadershipError, ConfigError
from warnings import warn
from typing import List
from re import compile, findall

from .convert import RougailConvert
from .config import RougailConfig
from .update import RougailUpgrade
from .object_model import CONVERT_OPTION
from .utils import normalize_family


def tiramisu_display_name(
    kls,
    subconfig,
    with_quote: bool = False,
) -> str:
    """Replace the Tiramisu display_name function to display path + description"""
    doc = kls._get_information(subconfig, "doc", None)
    comment = f" ({doc})" if doc and doc != kls.impl_getname() else ""
    if "{{ identifier }}" in comment:
        comment = comment.replace("{{ identifier }}", str(subconfig.identifiers[-1]))
    path = kls.impl_getpath()
    if "{{ identifier }}" in path and subconfig.identifiers:
        path = path.replace(
            "{{ identifier }}", normalize_family(str(subconfig.identifiers[-1]))
        )
    if with_quote:
        return f'"{path}"{comment}'
    return f"{path}{comment}"


class Rougail:
    """Main Rougail object"""

    def __init__(
        self,
        rougailconfig=None,
    ) -> None:
        if rougailconfig is None:
            rougailconfig = RougailConfig
        self.rougailconfig = rougailconfig
        self.converted = RougailConvert(self.rougailconfig)
        self.config = None

    def add_path_prefix(
        self,
        path_prefix: str,
    ) -> None:
        """Add a prefix"""
        self.converted.load_config()
        self.converted.parse_directories(path_prefix)

    def run(self):
        """Get Tiramisu Config"""
        if not self.config:
            tiram_obj = self.converted.save(self.rougailconfig["tiramisu_cache"])
            optiondescription = {}
            custom_types = {
                custom.__name__: custom
                for custom in self.rougailconfig["custom_types"].values()
            }
            exec(tiram_obj, custom_types, optiondescription)  # pylint: disable=W0122
            self.config = Config(
                optiondescription["option_0"],
                display_name=tiramisu_display_name,
            )
            self.config.property.read_write()
        return self.config

    def get_config(self):
        warn(
            "get_config is deprecated, use run instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.run()

    def user_datas(self, user_datas: List[dict]):
        values = {}
        errors = []
        warnings = []
        for datas in user_datas:
            options = datas.get("options", {})
            for name, data in datas.get("values", {}).items():
                values[name] = {
                    "values": data,
                    "options": options.copy(),
                }
            errors.extend(datas.get("errors", []))
            warnings.extend(datas.get("warnings", []))
        self._auto_configure_dynamics(values)
        while values:
            value_is_set = False
            for option in self._get_variable(self.config):
                path = option.path()
                if path not in values:
                    path = path.upper()
                    options = values.get(path, {}).get("options", {})
                    if path not in values or options.get("upper") is not True:
                        continue
                else:
                    options = values[path].get("options", {})
                value = values[path]["values"]
                if option.ismulti():
                    if options.get("multi_separator") and not isinstance(value, list):
                        value = value.split(options["multi_separator"])
                        values[path]["values"] = value
                    if options.get("needs_convert"):
                        value = [convert_value(option, val) for val in value]
                        values[path]["values"] = value
                        values[path]["options"]["needs_convert"] = False
                elif options.get("needs_convert"):
                    value = convert_value(option, value)
                index = option.index()
                if index is not None:
                    if not isinstance(value, list) or index >= len(value):
                        continue
                    value = value[index]
                try:
                    option.value.set(value)
                    value_is_set = True
                    if index is not None:
                        values[path]["values"][index] = undefined
                        if set(values[path]["values"]) == {undefined}:
                            values.pop(path)
                    else:
                        values.pop(path)
                except Exception as err:
                    if path != option.path():
                        values[option.path()] = values.pop(path)
            if not value_is_set:
                break
        for path, data in values.items():
            try:
                option = self.config.option(path)
                value = data["values"]
                if option.isfollower():
                    for index, val in enumerate(value):
                        if val is undefined:
                            continue
                        self.config.option(path, index).value.set(val)
                else:
                    option.value.set(value)
            except AttributeError as err:
                errors.append(str(err))
            except (ValueError, LeadershipError) as err:
                # errors.append(str(err).replace('"', "'"))
                errors.append(str(err))
            except PropertiesOptionError as err:
                #                warnings.append(f'"{err}" but is defined in "{self.filename}"')
                warnings.append(str(err))
        return {
            "errors": errors,
            "warnings": warnings,
        }

    def _get_variable(self, config):
        for subconfig in config:
            if subconfig.isoptiondescription():
                yield from self._get_variable(subconfig)
            else:
                yield subconfig

    def _auto_configure_dynamics(
        self,
        values,
    ):
        cache = {}
        added = []
        for path, data in list(values.items()):
            value = data["values"]
            #            for value in data['values'].items():
            try:
                option = self.config.option(path)
                option.name()
            except (ConfigError, PropertiesOptionError):
                pass
            except AttributeError:
                config = self.config
                current_path = ""
                identifiers = []
                for name in path.split(".")[:-1]:
                    if current_path:
                        current_path += "."
                    current_path += name
                    if current_path in cache:
                        config, identifier = cache[current_path]
                        identifiers.append(identifier)
                    else:
                        tconfig = config.option(name)
                        try:
                            tconfig.group_type()
                            config = tconfig
                        except AttributeError:
                            for tconfig in config.list(uncalculated=True):
                                if tconfig.isdynamic(only_self=True):
                                    identifier = self._get_identifier(
                                        tconfig.name(), name
                                    )
                                    if identifier is None:
                                        continue
                                    dynamic_variable = tconfig.information.get(
                                        "dynamic_variable",
                                        None,
                                    )
                                    if not dynamic_variable:
                                        continue
                                    option_type = self.config.option(
                                        dynamic_variable
                                    ).information.get("type")
                                    if identifiers:
                                        for s in identifiers:
                                            dynamic_variable = dynamic_variable.replace(
                                                "{{ identifier }}", str(s), 1
                                            )
                                    if dynamic_variable not in values:
                                        values[dynamic_variable] = {"values": []}
                                        added.append(dynamic_variable)
                                    elif dynamic_variable not in added:
                                        continue
                                    config = tconfig
                                    #                                        option_type = option.information.get('type')
                                    typ = CONVERT_OPTION.get(option_type, {}).get(
                                        "func"
                                    )
                                    if typ:
                                        identifier = typ(identifier)
                                    if (
                                        identifier
                                        not in values[dynamic_variable]["values"]
                                    ):
                                        values[dynamic_variable]["values"].append(
                                            identifier
                                        )
                                    identifiers.append(identifier)
                                    cache[current_path] = config, identifier
                                    break
            else:
                if option.isdynamic():
                    parent_option = self.config.option(path.rsplit(".", 1)[0])
                    identifiers = self._get_identifier(
                        parent_option.name(uncalculated=True),
                        parent_option.name(),
                    )
                    dynamic_variable = None
                    while True:
                        dynamic_variable = parent_option.information.get(
                            "dynamic_variable",
                            None,
                        )
                        if dynamic_variable:
                            break
                        parent_option = self.config.option(
                            parent_option.path().rsplit(".", 1)[0]
                        )
                        if "." not in parent_option.path():
                            parent_option = None
                            break
                    if not parent_option:
                        continue
                    identifiers = parent_option.identifiers()
                    for identifier in identifiers:
                        dynamic_variable = dynamic_variable.replace(
                            "{{ identifier }}", str(identifier), 1
                        )
                    if dynamic_variable not in values:
                        values[dynamic_variable] = {"values": []}
                        added.append(dynamic_variable)
                    elif dynamic_variable not in added:
                        continue
                    option_type = option.information.get("type")
                    typ = CONVERT_OPTION.get(option_type, {}).get("func")
                    if typ:
                        identifier = typ(identifier)
                    if identifier not in values[dynamic_variable]["values"]:
                        values[dynamic_variable]["values"].append(identifier)
                    cache[option.path()] = option, identifier

    def _get_identifier(self, true_name, name) -> str:
        regexp = true_name.replace("{{ identifier }}", "(.*)")
        finded = findall(regexp, name)
        if len(finded) != 1 or not finded[0]:
            return
        return finded[0]


def convert_value(option, value):
    if value == "":
        return None
    option_type = option.information.get("type")
    func = CONVERT_OPTION.get(option_type, {}).get("func")
    if func:
        return func(value)
    return value


__all__ = ("Rougail", "RougailConfig", "RougailUpgrade")



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_2 = StrOption(name="var1", doc="the first variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string', 'test': ('test',)})
option_3 = StrOption(name="var2", doc="the second variable", default="value", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string', 'test': ('test',)})
option_4 = StrOption(name="var3", doc="the third variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string', 'test': ('test1', 'test2')})
option_5 = StrOption(name="var4", doc="the forth variable", properties=frozenset({"standard"}), informations={'type': 'string', 'test': (None, 'test1', 'test2')})
option_6 = BoolOption(name="var5", doc="the fifth variable", default=True, properties=frozenset({"mandatory", "standard"}), informations={'type': 'boolean', 'test': (False,)})
option_7 = StrOption(name="var6", doc="the sixth variable", multi=True, properties=frozenset({"basic", "mandatory"}), informations={'type': 'string', 'test': ('test1', 'test2')})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, option_3, option_4, option_5, option_6, option_7], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_1 = IPOption(name="nut_monitor_host", doc="nut_monitor_host", default="192.168.0.1", allow_reserved=True, properties=frozenset({"mandatory", "normal"}))
optiondescription_16 = OptionDescription(name="rougail", doc="Rougail", children=[option_1], properties=frozenset({"normal"}))
option_7 = SymLinkOption(name="name", opt=option_1)
option_6 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_5 = OptionDescription(name="nut_monitor_host", doc="nut_monitor_host", children=[option_7, option_6])
optiondescription_4 = OptionDescription(name="ip", doc="ip", children=[optiondescription_5])
option_3 = BoolOption(name="activate", doc="activate", default=True)
option_8 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_18 = OptionDescription(name="nut_service", doc="nut.service", children=[optiondescription_4, option_3, option_8])
optiondescription_18.impl_set_information('type', "service")
optiondescription_17 = OptionDescription(name="services", doc="services", children=[optiondescription_18], properties=frozenset({"hidden"}))
optiondescription_15 = OptionDescription(name="1", doc="1", children=[optiondescription_16, optiondescription_17])
option_2 = IPOption(name="nut_monitor_host", doc="nut_monitor_host", default="192.168.0.1", allow_reserved=True, properties=frozenset({"mandatory", "normal"}))
optiondescription_20 = OptionDescription(name="rougail", doc="Rougail", children=[option_2], properties=frozenset({"normal"}))
option_13 = SymLinkOption(name="name", opt=option_2)
option_12 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_11 = OptionDescription(name="nut_monitor_host", doc="nut_monitor_host", children=[option_13, option_12])
optiondescription_10 = OptionDescription(name="ip", doc="ip", children=[optiondescription_11])
option_9 = BoolOption(name="activate", doc="activate", default=True)
option_14 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_22 = OptionDescription(name="nut_service", doc="nut.service", children=[optiondescription_10, option_9, option_14])
optiondescription_22.impl_set_information('type', "service")
optiondescription_21 = OptionDescription(name="services", doc="services", children=[optiondescription_22], properties=frozenset({"hidden"}))
optiondescription_19 = OptionDescription(name="2", doc="2", children=[optiondescription_20, optiondescription_21])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_15, optiondescription_19])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_3 = StrOption(name="var2", doc="a second variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_4 = StrOption(name="var3", doc="a third variable", default="yes", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_2 = StrOption(name="var", doc="a first variable", multi=True, default=[Calculation(func['calc_value'], Params((ParamOption(option_3)))), Calculation(func['calc_value'], Params((ParamOption(option_4))))], default_multi=Calculation(func['calc_value'], Params((ParamOption(option_3)))), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, option_3, option_4], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])
